/* test of points in a ball */
#include <stdlib.h>
#include <stdio.h>			// C I/O
#include <stdint.h>
#include <fstream>			// file I/O
#include <iostream>			// file I/O
#include <string.h>			// string manipulation
#include <math.h>			// math routines
#include <list>
#include <queue>
#include <time.h>

#include <DNN/ANN.h>			// ANN declarations
#include <DNN/multiann.h>

#ifndef INFINITY
#define INFINITY 1.0e40
#endif

#ifndef PI
#define PI 3.1415926535897932385
#endif

using namespace std;

unsigned int    numTest         = 1;            // Number of test (repetitions)
double          R               = 1.5;          // Radius of search
int		MaxNeighbors    = 1;		// number of nearest neighbors (not used)
int		dim		= 15;		// dimension
int		m_pts		= 50000;       	// maximum number of data points
int		numIt		= 1000;       	// maximum number of nearest neighbor calls

istream		*dim_in		= NULL;		// input for dimension
istream		*topology_in	= NULL;		// input for topology

void randomDistPoint(int dim, int *tt, ANNpoint &p)
{
    for (int kk = 0; kk < dim; kk++){
      double tmp = (double)rand()/(double)RAND_MAX;

      if (tt[kk] == 2) {
        p[kk] = (2*PI*tmp - PI); // region -PI PI
      }
      else if (tt[kk] == 3) {
	double tmp1 = rand()/(2.14783e+09);
	double tmp2 = rand()/(2.14783e+09);
	
	double Theta1 = 2*PI*tmp1;
	double Theta2 = PI*tmp2 - PI/2;
	double r1 = sqrt(1 - tmp);
	double r2 = sqrt(tmp);
	
	p[kk] =  sin(Theta1)*r1;
	p[kk + 1] = cos(Theta1)*r1;
	p[kk + 2] = sin(Theta2)*r2;
	p[kk + 3] = cos(Theta2)*r2;
	
	kk = kk + 3;		// random quaternions
      }
      else 
	p[kk] = 2*tmp-1; 	//region -1 to 1
    }

}

void randomDist(int dim, int *tt, ANNpointArray &data_pts) 
{
  int n_pts = 0;
  while (n_pts < m_pts) {
    randomDistPoint(dim, tt, data_pts[n_pts]);
    n_pts++;
  }
}

// Default is to use the standard Euclidean metric
double Metric(ANNpoint x1, ANNpoint x2, int dim, int *topology, ANNpoint scale) {

  double rho = 0, fd, dtheta;
  
  for (int i = 0; i < dim; i++) {
    if (topology[i] == 1) {
      rho += ANN_POW(scale[i]*(x1[i] - x2[i]));
    }
    else if (topology[i] == 2) {
      fd = fabs(x1[i] - x2[i]);
      dtheta = ANN_MIN(fd, 2.0*PI - fd);
      rho += ANN_POW(scale[i]*dtheta);
    }
    else if (topology[i] == 3) {
      fd = x1[i]*x2[i] + x1[i+1]*x2[i+1] + x1[i+2]*x2[i+2] + x1[i+3]*x2[i+3];
      if (fd > 1) {
	double norm1 = x1[i]*x1[i] + x1[i+1]*x1[i+1] + x1[i+2]*x1[i+2] + x1[i+3]*x1[i+3];
	double norm2 = x2[i]*x2[i] + x2[i+1]*x2[i+1] + x2[i+2]*x2[i+2] + x2[i+3]*x2[i+3];
	fd = fd/(sqrt(norm1*norm2));
      }
      dtheta = ANN_MIN(acos(fd), acos(-fd));
      rho += ANN_POW(scale[i]*dtheta);
      i = i + 3;
    }
  }
  //cout << "dist = " << sqrt(rho) << endl <<endl;

  return sqrt(rho);
}

ANNbool readPt(istream &in, int *p)		// read point (false on EOF)
{
    for (int i = 0; i < dim; i++) {
	if(!(in >> p[i])) return ANNfalse;
    }
    return ANNtrue;
}

void printPt(ostream &out, ANNpoint p)		// print point
{
    out << "(" << p[0];
    for (int i = 1; i < dim; i++) {
	out << ", " << p[i];
    }
    out << ")\n";}

int main(int argc, char **argv)
{
  int k_ann;
  double d_brute;
  double *d_ann;
  int *idx_ann;

    static ifstream dimStream;			// dimension file stream
    static ifstream topologyStream;		// topology file stream

    ANNpointArray	data_pts;		// data points
    ANNpoint		query_pt;		// query point
    int 		*topology;		// topology of the space
    ANNpoint		scale;			// scaling of the coordinates
    ANNpoint		result_pt;     		// scaling of the coordinates
    MultiANN		*MAG;			// search structure


    srand(time(NULL)); 
    dimStream.open("test/dim", ios::in);		// open query file
    if (!topologyStream) {
      cerr << "Cannot open dim file\n";
      exit(1);
    }
    dim_in = &dimStream;			// make this query stream
     (*dim_in) >> dim;				// read dimension


    query_pt = annAllocPt(dim);			// allocate query point
    data_pts = annAllocPts(m_pts, dim);		// allocate data points
    topology = new int[dim];			// allocate topology array
    scale = new double[dim];			// allocate scaling array


    topologyStream.open("test/topology", ios::in);	// open query file
    if (!topologyStream) {
      cerr << "Cannot open topology file\n";
      exit(1);
    }
    topology_in = &topologyStream;	      	// make this query stream
    readPt(*topology_in, topology);		// read topology
    
    for (int i = 0; i < dim; i++) {
      if (topology[i] == 1) scale[i] = 1.0;
      else if (topology[i] == 2) scale[i] = 50/PI;
      else if (topology[i] == 3) scale[i] = 50.0;
    }
    cout << "Topology and Scale: \n";			//print topology
    for (int i = 0; i < dim; i++) {
      cout << " topology[" << i << "] = " << topology[i] << "\t";
      cout << " scale[" << i << "] = " << scale[i] << endl;
    }
    cout << "\n";

    randomDist(dim, topology, data_pts);		// generate random points
    randomDistPoint(dim, topology, query_pt);		// generate query point
    
    cout << "R=" << R << "\t"; 
    cout << "m_pts=" << m_pts << "\t";
    cout << "dim=" << dim << "\t";
    cout << "numTest=" << numTest << "\t";
    cout << "numIt=" << numIt << "\n";

  for (int ii = 0; ii < numTest; ii++) {

    cout << "\nTest " << ii << ":" << endl;

    randomDist(dim, topology, data_pts);		// generate random points
    randomDistPoint(dim, topology, query_pt);		// generate query point

    //***********************Construction phase**********************/
    
    clock_t tv1, tv2;
    double time;    
    tv1 = clock();

    MAG = new MultiANN(dim, MaxNeighbors, topology, scale);	// create a data structure	
    for (int j = 0; j < m_pts; j++) {
      MAG->AddPoint(data_pts[j], data_pts[j]);		// add new data point
    }

    tv2 = clock();
    time = (tv2 - tv1)/(CLOCKS_PER_SEC / (double) 1000.0);
    cout << "  Construction MPNN time: "<< time << " ms\n";
    tv1 = clock();

    //*************************Query phase*******************************/
    
    
    // compute nearest neighbor using library
    for (int j = 0; j < numIt; j++) {
      randomDistPoint(dim, topology, query_pt);				// generate query point
      MAG->NearestNeighbor(query_pt,R,&k_ann,&d_ann,&idx_ann); 
      if (k_ann>0)
	{
	  delete [] idx_ann;
	  delete [] d_ann;
	}
    } 
    tv2 = clock();
    time = (tv2 - tv1)/(CLOCKS_PER_SEC / (double) 1000.0);
    cout << "  MPNN query time: "<< time/(double)numIt << " ms\n";

  

    tv1 = clock();

    // compare the obtained result with the brute force result
    for (int j = 0; j < numIt; j++) {
      randomDistPoint(dim, topology, query_pt);				// generate query point
      for (int i = 0; i < m_pts; i++) {
	d_brute = Metric(query_pt, data_pts[i], dim, topology, scale);
      }
    }
    tv2 = clock();
    time = (tv2 - tv1)/(CLOCKS_PER_SEC / (double) 1000.0);
    cout << "  Brute force time: "<< time/(double)numIt << "ms\n";

    if (MAG){
      MAG->ResetMultiANN();
      MAG = NULL;
    }
    
  }  
}






